module Bfs
 (
  distances, path, rPath, reversePath,
  closeDistances, closePath
 ) where

import IArray
import MArray
import ST
import Monad

import Data
import Utils

type Queue a = ([a],[a])

emptyQueue::Queue a
emptyQueue=([],[])

isEmptyQueue::Queue a->Bool
isEmptyQueue ([],[]) = True
isEmptyQueue _ = False

shift::Queue a->a->Queue a
shift (i,o) x = (x:i,o)

unshift::Queue a->(a,Queue a)
unshift (i,[]) = unshift ([],reverse i)
unshift (i,h:t) = (h,(i,t))

bfs::Array Point Field->Queue (Point,Int)->STUArray s Point Int->ST s ()
bfs mp q dists
 | isEmptyQueue q = return ()
 | otherwise =
    do
     let nb = filter isFree $ map (move a) dirList
     nb' <- filterM wasNot nb
     mapM_ (\p->writeArray dists p (d+1)) nb'
     let q' = foldl shift rest $ map (\x->(x,d+1)) nb'
     bfs mp q' dists
 where
  ((a,d),rest) = unshift q
  isFree pos = mp ! pos == FEmpty
  wasNot pos = do
                d <- readArray dists pos
                return (d == (-1))

limBfs::Array Point Field->Queue (Point,Int)->STUArray s Point Int->Int->ST s ()
limBfs mp q dists range
 | isEmptyQueue q = return ()
 | d == range = return ()
 | otherwise =
    do
     let nb = filter isFree $ map (move a) dirList
     nb' <- filterM wasNot nb
     mapM_ (\p->writeArray dists p (d+1)) nb'
     let q' = foldl shift rest $ map (\x->(x,d+1)) nb'
     limBfs mp q' dists range
 where
  ((a,d),rest) = unshift q
  isFree pos = mp ! pos == FEmpty
  wasNot pos = do
                d <- readArray dists pos
                return (d == (-1))

distancesST::Board->Point->ST s (STUArray s Point Int)
distancesST Board {bW=w,bH=h,bC=mp} from =
 do
  dists <- newArray ((0,0),(w+1,h+1)) (-1)
  let q = shift emptyQueue (from,0)
  writeArray dists from 0
  bfs mp q dists
  return dists

closeDistancesST::Board->Point->Int->ST s (STUArray s Point Int)
closeDistancesST board from range =
 do
  let rng = (range+1,range+1)
  dists <- newArray (from `subPoint` rng, from `addPoint` rng) (-1)
  let q = shift emptyQueue (from,0)
  writeArray dists from 0
  limBfs (bC board) q dists range
  return dists

distances::Board->Point->UArray Point Int
distances board from = runST (distancesST board from >>= unsafeFreeze)

closeDistances::Board->Point->Int->UArray Point Int
closeDistances board from range =
  runST (closeDistancesST board from range >>= unsafeFreeze)

closePath::UArray Point Int->Point->Point->Int->Maybe [Dir]
closePath surrDistances from to range
 | vic from to > range = Nothing
 | surrDistances ! to == -1 = Nothing
 | otherwise = Just (path surrDistances to)

rPath::UArray Point Int->Point->[Dir]
rPath dists pos = path' pos
 where
  path' pos
   | d == 0 = []
   | otherwise = dir : path' nPos
   where
    d = dists ! pos
    nb = zip dirList $
         map (\p -> (p,dists ! p)) $
         map (move pos) dirList
    (dir,(nPos,_)) = head $ filter (\(_,(_,dd)) -> dd == d - 1) nb

reversePath::[Dir]->[Dir]
reversePath = reverse . map rDir
 where
  rDir DNorth = DSouth
  rDir DSouth = DNorth
  rDir DEast = DWest
  rDir DWest = DEast

path::UArray Point Int->Point->[Dir]
path dists from = reversePath $ rPath dists from

