module Checks where

import FiniteMap
import Set
import Maybe

import Data
import Utils

andCheck::Check->Check->Check
andCheck c1 c2 board self state = c1 board self state && c2 board self state

notCheck::Check->Check
notCheck c board self state = not (c board self state)

alwaysOK::Check
alwaysOK _ _ _ = True

alwaysWrong::Check
alwaysWrong _ _ _ = False

checkAt::Point->Check
checkAt pos board self state = pos == rPos (findMe state self)

checkUnpickable::[Int]->Check
checkUnpickable ids board self state = and $ map noPick ids
 where
  me = findMe state self
  noPick i
   | Right _ <- pPos pack = True
   | Left pos <- pPos pack = pos /= rPos me
   where
    pack = findPackage state i

checkDelivered::[Int]->Check
checkDelivered l board self state = and $ map checkDeliver l
 where
  packages = sPackages state
  checkDeliver pack = not (elemFM pack packages)

checkHave::Int->Check
checkHave pack board self state = elementOf pack $ rPackages me
 where
  me = findMe state self

checkPackageAt::Int->Point->Check
checkPackageAt pack pos board self state
 | Nothing <- pkg' = False
 | otherwise = pPos pkg == Left pos
 where
  packages = sPackages state
  pkg' = lookupFM packages pack
  Just pkg = pkg'

checkAlive::Int->Check
checkAlive r board self state = isJust robot
 where
  robots = sRobots state
  robot = lookupFM robots r

checkHoldsPackage::Int->Check
checkHoldsPackage r board self state
 | Nothing <- robot = False
 | Just rob <- robot = not (isEmptySet $ rPackages rob)
 where
  robots = sRobots state
  robot = lookupFM robots r

checkFreePackages::Check
checkFreePackages board self state = not (isEmptyFM free)
 where
  packages = sPackages state
  bases = sBases state
  free = filterFM (\_ -> testWeight self) $ filterFM (\_ -> testDestination board) $
         filterFM (\_ pack -> isLeft (pPos pack)) packages

checkBaseExists::Point->Check
checkBaseExists base board self state = elementOf base (sBases state)

