module Data where

import FiniteMap
import Set
import Array

type Point = (Int,Int)

data Robot = Robot {
                    rPos::Point,
                    rPackages::Set Int
                   }

data Package = Package {
                        pDest::Maybe Point,
                        pWeight::Maybe Int,
                        pPos::Either Point Int,
                        pSure::Bool
                       }

data State = State {
                    sBases::Set Point,
                    sRobots::FiniteMap Int Robot,
                    sPackages::FiniteMap Int Package
                   }

data Field = FEmpty |
             FWater |
             FWall |
             FUnreachable
             deriving (Eq,Show)

data Board = Board {
                    bW, bH :: Int,
                    bC :: Array Point Field
                   }

data Handler =
  forall x . Handler (Board->Self->State->Maybe x) (x->Strategy) Bool

type Check = Board->Self->State->Bool
data Strategy = Strategy {
                          strName::String,
                          strGenerate::Board->Self->State->Maybe [Move],
                          strCheck::Board->Self->State->Move->Check,
                          strHandlers::[Handler]
                         }

data Self = Self {
                  sId, sCapa, sMoney :: Int,
                  sActions :: [(Check,[Move],Strategy)]
                 }

data Dir = DNorth |
           DEast |
           DSouth |
           DWest deriving (Eq,Ord,Ix)

instance Show Dir where
 showsPrec _ DNorth = showString "N"
 showsPrec _ DEast = showString "E"
 showsPrec _ DSouth = showString "S"
 showsPrec _ DWest = showString "W"

dirs::Array Dir Point
dirs = array (DNorth,DWest) [(DNorth,(0,1)),
                             (DEast,(1,0)),
                             (DSouth,(0,-1)),
                             (DWest,(-1,0))]

dirList::[Dir]
dirList = [DNorth,DEast,DSouth,DWest]

move::Point->Dir->Point
move (x,y) dir = (x+dx,y+dy)
 where
  (dx,dy) = dirs ! dir

data RMove = MMove Dir |
             MDrop [Int] |
             MPick [Int]

data Move = Move Int RMove

instance Show Move where
 showsPrec _ (Move c m) = shows c . showChar ' ' . sh m
  where
   sh (MMove dir) = showString "Move " . shows dir
   sh (MDrop ids) = showString "Drop" . shids ids
   sh (MPick ids) = showString "Pick" . shids ids
   shids [] = id
   shids (h:t) = showChar ' ' . shows h . shids t

