module Main where

import System
import IO
import Socket
import IArray
import FiniteMap
import Set
import Maybe

import Data
import Utils
import Checks(alwaysOK,alwaysWrong)
import Strategies
import Update
import Bfs

decide::Board->Self->State->IO (Self,Move)
decide board self state
 | null actions =
     do
      putStrLn "Applying default strategy"
      decide board self{sActions=[(alwaysWrong,[],defaultStrategy board self state)]} state
 | (ok || isJust newMoves) && not (null appHandlers) =
     do
      putStrLn "Handler activated"
      decide board hanSelf state
 | ok && null moves =
     do
      putStrLn $ "Strategy " ++ show (strName strategy) ++ " succeeded"
      decide board self {sActions = rest} state
 | ok = do
         let move = head moves
             rmoves = tail moves
             check' = strCheck strategy board self state move
         return (self {sActions=(check',rmoves,strategy):rest},move)
 | Just moves' <- newMoves =
        do
         putStrLn $ "Strategy " ++ show (strName strategy) ++ " resumed"
         decide board self{sActions = (alwaysOK,moves',strategy):rest} state
 | otherwise =
        do
         putStrLn $ "Strategy " ++ show (strName strategy) ++ " failed"
         decide board self{sActions = rest} state
 where
  newMoves = strGenerate strategy board self state
  actions = sActions self
  ((check,moves,strategy):rest) = actions
  ok = check board self state
  appHandlers = mapMaybe appHandler (strHandlers strategy)
  appHandler (Handler app str inter) = fmap (\res->(str res,inter)) $ app board self state
  hanSelf
   | ((str,False):_) <- appHandlers = self{sActions = (alwaysWrong,[],str) : actions}
   | ((str,True):_) <- appHandlers = self{sActions = (alwaysWrong,[],str) : rest}

processMove::Move->Self->Self
processMove (Move c _) s = s {sMoney = sMoney s - c}

dumpState::Board->Self->State->IO ()
dumpState board self state =
 do
  putStrLn "------------------------------------------------------------------"
  let mp = drawMap board
      mp1 = putBases mp (setToList $ sBases state)
      mp2 = putRobots mp1 (fmToList $ sRobots state) (sId self)
  mapM_ putStrLn mp2
  putStrLn ""
  mapM_ (putStrLn . drawPackage) $ fmToList $ sPackages state
  putStrLn ""
  putStrLn $ "Fund: " ++ show (sMoney self)
  putStrLn $ "Capacity: " ++ show (capa state self) ++ " (" ++ show (sCapa self) ++ ")"
  putStrLn "------------------------------------------------------------------"
 where
  c = bC board
  drawMap board = [ [sField (c ! (x,y)) | x <- [0..bW board + 1]] | y <- [0..bH board + 1]]
  sField FEmpty = '.'
  sField FWater = '~'
  sField FWall = '#'
  sField FUnreachable = '+'
  putBases mp l = foldl (draw '@') mp l
  putRobots mp [] _ = mp
  putRobots mp ((i,r):t) si
   | i == si = draw 'X' mp' (rPos r)
   | otherwise = draw 'o' mp' (rPos r)
   where
    mp' = putRobots mp t si
  draw c mp (x,y) = take y mp ++ [drawL c (mp !! y) x] ++ drop (y+1) mp
  drawL c l x = take x l ++ [c] ++ drop (x+1) l
  drawPackage (i,pack) = "Package " ++ show i ++ ": pos " ++ show (pPos pack) ++
                         (if pSure pack then "(sure)" else "") ++
                         (if isJust (pDest pack) then " dest " ++ show (fromJust $ pDest pack) else "") ++
                         (if isJust (pWeight pack) then " weight " ++ show (fromJust $ pWeight pack) else "")

play::Handle->Board->Self->State->IO ()
play h board self state =
 do
  -- dumpState board self state
  (self',move) <- decide board self state
  let newSelf = processMove move self'
  hPutStrLn h $ {-debug $-} show move
  newState <- updateState h board (sId newSelf) state -- process status line & packages
  play h board newSelf newState

playFirst::Handle->Board->Self->State->IO ()
playFirst h board self state =
 do
  newState <- updateState h board (sId self) state -- process status line & packages
  let dists = distances board (rPos $ findMe newState self)
      wg = bW board
      hg = bH board
      c = bC board
      c' = map rAble [(x,y)| x<-[0 .. wg+1],y <- [0 .. hg+1]]
      rAble pos
       | c ! pos == FEmpty && dists ! pos == -1 = (pos,FUnreachable)
       | otherwise = (pos, c ! pos)
      newC = array ((0,0),(wg+1,hg+1)) c'
      board' = board {bC = newC}
      dead = filter (\(i,r)->newC ! rPos r == FUnreachable) $ fmToList $ sRobots newState
      resState = killDeadRobots dead newState
      bases = sBases state
      newBases = mkSet $ filter (\p->newC ! p /= FUnreachable) $ setToList bases
  play h board' self resState{sBases=newBases}

main::IO ()
main=
 do
  (addr:port:_) <- getArgs
  h <- connectTo addr (PortNumber $ fromInteger $ read port)
  hSetBuffering h LineBuffering
  hPutStrLn h "Player"
  (board,bases) <- getBoard h
  self <- getSelf h
  catch (playFirst h board self State {sRobots=emptyFM,sPackages=emptyFM,sBases=mkSet bases}) (\_->return ())
  rest <- hGetContents h
  putStr rest
  return ()

