module Update
 (
  getBoard, getSelf, updateState, killDeadRobots
 ) where

import IO
import FiniteMap
import Set
import Array
import Maybe

import Data

getBoard::Handle->IO (Board,[Point])
getBoard h =
 do
  wh <- hGetLine h
  let [sw,sh] = words wh
      wg = read sw
      hg = read sh
  mp <- sequence $ replicate hg $ hGetLine h
  return (Board {bW=wg,bH=hg,bC = array ((0,0),(wg+1,hg+1)) (board wg hg mp)},
          bases wg mp)
 where
  bases w mp = concat (zipWith (lBases w) [1..] mp)
  lBases w y l = mapMaybe (recBase y) (zip [1..] $ take w l)
  recBase y (x,'@') = Just (x,y)
  recBase _ _ = Nothing
  board w h mp = wallLine 0 (w+1) ++ concat (zipWith (xLine w) [1..] mp) ++ wallLine (h+1) (w+1)
  wallLine y w = map (\x->((x,y),FWall)) [0..w]
  xLine w y l = ((0,y),FWall) : zipWith (rField y) [1..] (take w l) ++ [((w+1,y),FWall)]
  rField y x ch = ((x,y),iCh ch)
  iCh '.' = FEmpty
  iCh '~' = FWater
  iCh '#' = FWall
  iCh '@' = FEmpty

getSelf::Handle->IO Self
getSelf h =
 do
  l <- hGetLine h
  let [sId,sCapa,sMoney] = words l
  return Self {sActions = [], sId = read sId, sCapa = read sCapa, sMoney = read sMoney}

killDeadRobot::(Int,Robot)->State->State
killDeadRobot (i,r) s@State {sRobots = robots, sPackages = packages} =
  s {sRobots = delFromFM robots i,
    sPackages = delListFromFM packages $ setToList $ rPackages r}

updateMove::Int->Int->Dir->State->State
updateMove selfId i dir state
  | Nothing <- mRobot = state
  | otherwise = state {sRobots=robots',sBases=bases'}
 where
  robots = sRobots state
  bases = sBases state
  mRobot = lookupFM robots i
  Just robot = mRobot
  newPos = rPos robot `move` dir
  bases' = if selfId /= i then bases else delFromSet bases newPos
  robots' = addToFM robots i (robot {rPos = newPos})

updateDrop::Int->Int->State->State
updateDrop i pack state
 | Nothing <- mRobot = state
 | Just pos <- pDest package, pos == rPos robot =
      state {sRobots=robots', sPackages = dPackages}
 | otherwise =
      state {sRobots=robots', sPackages = packages'}
 where
  robots = sRobots state
  packages = sPackages state
  mRobot = lookupFM robots i
  Just robot = mRobot
  package = lookupWithDefaultFM packages Package {pDest=Nothing, pWeight=Nothing} pack
  robot' = robot {rPackages = rPackages robot `delFromSet` pack}
  package' = package {pPos=Left (rPos robot),pSure=isJust (pDest package)}
  robots' = addToFM robots i robot'
  packages' = addToFM packages pack package'
  dPackages = delFromFM packages pack

updatePick::Int->Int->State->State
updatePick i pack state
 | Nothing <- mRobot = state
 | otherwise = state {sRobots=robots',sPackages = packages'}
 where
  robots = sRobots state
  packages = sPackages state
  mRobot = lookupFM robots i
  Just robot = mRobot
  package = lookupWithDefaultFM packages Package {pDest=Nothing, pWeight=Nothing} pack
  robot' = robot {rPackages = rPackages robot `addToSet` pack}
  package' = package {pPos=Right i,pSure=True}
  robots' = addToFM robots i robot'
  packages' = addToFM packages pack package'

updatePos::Board->Int->Point->State->State
updatePos board i pnt state
 | bC board ! pnt == FUnreachable = state
 | otherwise = state{sRobots=robots'}
 where
  robots = sRobots state
  robot = lookupWithDefaultFM robots (Robot {rPackages=emptySet}) i
  robots' = addToFM robots i robot{rPos=pnt}

updateRobot::Board->Int->(Int,[String])->State->State
updateRobot _ _ (_,[]) = id
updateRobot b selfId (i,"N":rest) = updateRobot b selfId (i,rest) . updateMove selfId i DNorth
updateRobot b selfId (i,"S":rest) = updateRobot b selfId (i,rest) . updateMove selfId i DSouth
updateRobot b selfId (i,"W":rest) = updateRobot b selfId (i,rest) . updateMove selfId i DWest
updateRobot b selfId (i,"E":rest) = updateRobot b selfId (i,rest) . updateMove selfId i DEast
updateRobot b selfId (i,"D":p:rest) = updateRobot b selfId (i,rest) . updateDrop i (read p)
updateRobot b selfId (i,"P":p:rest) = updateRobot b selfId (i,rest) . updatePick i (read p)
updateRobot b selfId (i,"X":x:"Y":y:rest) = updateRobot b selfId (i,rest) . updatePos b i (read x,read y)

killDeadRobots::[(Int,Robot)]->State->State
killDeadRobots r = foldl (.) id (map killDeadRobot r)

updateRobots::Board->String->Int->State->State
updateRobots board updates selfId state =
  foldl (.) id (map (updateRobot board selfId) rUpdates) state'
 where
  rUpdates = map (\x->(read $ head x, tail x)) $ map words $ splitUpdates updates
  living = map fst rUpdates
  dead = fmToList $ delListFromFM (sRobots state) living
  state' = killDeadRobots dead state

  splitUpdates [] = []
  splitUpdates ('#':rest) = a:splitUpdates r
   where
    (a,r) = span (/= '#') rest

updatePackages::String->Point->State->State
updatePackages packs pos = updatePackages' packs'
 where
  packs' = words packs
  removeCurrent state = state {sPackages = packages'}
   where
    packages = sPackages state
    here = filterFM (\_ pack -> pPos pack == Left pos) packages
    packages' = minusFM packages here
  updatePackages' [] = removeCurrent
  updatePackages' (i:dx:dy:w:rest) =
    updatePackage (read i) (read dx) (read dy) (read w) . updatePackages' rest
  updatePackage i dx dy w state = state {sPackages=packages'}
   where
    packages = sPackages state
    packages' = addToFM packages i Package {pDest = Just (dx,dy),
                                            pWeight = Just w,
                                            pPos = Left pos,
                                            pSure = True}

updateState::Handle->Board->Int->State->IO State
updateState h board i state =
 do
  changes <- hGetLine h
  packages <- hGetLine h
  let state' = updateRobots board changes i state
      Just self = lookupFM (sRobots state') i
  case words packages of
   ("Robot":_) -> ioError (userError "Konec")
   _           -> return $ updatePackages packages (rPos self) state'

